// Scintilla source code edit control
// Document.cc - manages the text of the document
// Copyright 1998-1999 by Neil Hodgson <neilh@hare.net.au>
// The License.txt file describes the conditions under which this software may be distributed.

#ifndef GTK
#include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "Document.h"

static void dprintf(char *szFormat, ...) {
#ifdef TRACE
	char szBuffer[1000];
	char *pArguments = (char *) & szFormat + sizeof(szFormat);
	wvsprintf(szBuffer,szFormat,pArguments);
	OutputDebugString(szBuffer);
#endif
}

LineCache::LineCache() {
	valid = false;
	cache = new int[growSize];
	size = growSize;
	lines = 0;
}

LineCache::~LineCache() {
	delete []cache;
	cache = 0;
}

void LineCache::SetValue(int pos, int value) {
	//dprintf("SetValue[%d] = %d\n", pos, value);
	if (pos >= size) {
		//dprintf("Resize %d %d\n", size,pos);
		int sizeNew = pos + growSize;
		int *cacheNew = new int[sizeNew];
		if (!cacheNew) {
			dprintf("No memory available\n");
		}
		for (int i=0;i<size;i++)
			cacheNew[i] = cache[i];
		delete []cache;
		cache = cacheNew;
		size = sizeNew;
		//dprintf("end Resize %d %d\n", size,pos);
	}
	cache[pos] = value;
}

int LineCache::LineFromPosition(int pos) {
	if (lines == 0)
		return 0;
	if (pos >= cache[lines])
		return lines - 1;
	int lower = 0;
	int upper = lines;
	int middle = 0;
	do {
		middle = (upper + lower + 1) / 2;	// Round high
		if (pos < cache[middle]) {
			upper = middle - 1;
		} else {
			lower = middle;
		}
	} while (lower < upper);
	return lower;
	//while (line < lines && pos >= cache[line])
	//	line++;
	//return line > 0 ? line-1 : line;
}

Action::Action() {
	at = startAction;
	position = 0;
	data = 0;
	lenData = 0;
}

Action::~Action() {
}

void Action::Destroy() {
	delete []data;
	data = 0;
}

void Action::Create(actionType at_, int position_, char *data_, int lenData_) {
	delete []data;
	position = position_;
	at = at_;
	data = data_;
	lenData = lenData_;
}

Document::Document(int initialLength) {
	body = new char[initialLength];
	size = initialLength;
	length = 0;
	part1len = 0;
	gaplen = initialLength;
	part2body = body + gaplen;

	actions = new Action[30000];
	lenActions = 30000;
	maxAction = 0;
	currentAction = 0;
	actions[currentAction].Create(startAction);

	readOnly = false;
}

Document::~Document() {
	delete []body;
	body = 0;

#ifdef NEED_UNDO_LOG
	FILE *fp = fopen("Log.log", "wt");
	fprintf(fp,"Max = %3d\n", maxAction);
	fprintf(fp,"Current = %3d\n", currentAction);
	for (int i=0;i<maxAction;i++) {
		fprintf(fp,"%3d ", i);
		if (actions[i].at == startAction) {
			fprintf(fp,"=");
		} else if (actions[i].at == removeAction) {
			fprintf(fp,"- %4d %4d ", actions[i].position, actions[i].lenData);
			for (int j=0; j<20&&j<actions[i].lenData;j++)
				fprintf(fp,"%c", actions[i].data[j]);
		} else {
			fprintf(fp,"+ %4d %4d ", actions[i].position, actions[i].lenData);
			for (int j=0; j<20&&j<actions[i].lenData;j++)
				fprintf(fp,"%c", actions[i].data[j]);
		}
		fprintf(fp,"\n");
	}
	fclose(fp);
#endif
}

void Document::GapTo(int position) {
	if (position == part1len)
		return;
	if (position < part1len) {
		int diff = part1len - position;
		dprintf("Move gap backwards to %d diff = %d part1len=%d length=%d \n", position,diff, part1len, length);
		for (int i=0;i<diff;i++)
			body[part1len + gaplen - i - 1] = body[part1len - i - 1];
	} else {	// position > part1len
		int diff = position - part1len;
		dprintf("Move gap forwards to %d diff =%d\n", position,diff);
		for (int i=0;i<diff;i++)
			body[part1len + i] = body[part1len + gaplen + i];
	}
	part1len = position;
	part2body = body + gaplen;
}

void Document::RoomFor(int insertionLength) {
	//dprintf("need room %d %d\n", gaplen, insertionLength);
	if (gaplen <= insertionLength) {
		//dprintf("need room %d %d\n", gaplen, insertionLength);
		GapTo(length);
		int newSize = size + insertionLength + 4000;
		//dprintf("moved gap %d\n", newSize);
		char *newBody = new char[newSize];
		memcpy(newBody, body, size);
		delete []body;
		body = newBody;
		gaplen += newSize - size;
		part2body = body + gaplen;
		size = newSize;
		//dprintf("end need room %d %d - size=%d length=%d\n", gaplen, insertionLength,size,length);
	}
}

char Document::ByteAt(int position) {

	if (position < 0) {
		dprintf("Bad position %d\n",position);
		return 1;
	}
	if (position >= length + 10) {
		dprintf("Very Bad position %d of %d\n",position,length);
		//char sz[30];
		//gets(sz);
		exit(3);
	}
	if (position >= length) {
		dprintf("Bad position %d of %d\n",position,length);
		//char sz[30];
		//gets(sz);
		return 2;
	}

	if (position < part1len) {
		return body[position];
	} else {
		return part2body[position];
	}
}

char Document::SetByteAt(int position, char ch) {

	if (position < 0) {
		dprintf("Bad position %d\n",position);
		return 1;
	}
	if (position >= length + 11) {
		dprintf("Very Bad position %d of %d\n",position,length);
		//exit(2);
		return 3;
	}
	if (position >= length) {
		dprintf("Bad position %d of %d\n",position,length);
		return 2;
	}

	if (position < part1len) {
		body[position] = ch;
	} else {
		part2body[position] = ch;
	}
	return 0;
}

char Document::CharAt(int position) {
	return ByteAt(position*2);
}

char Document::StyleAt(int position) {
	return ByteAt(position*2 + 1);
}

void Document::InsertString(int position, char *s, int insertLength) {
	// InsertString and DeleteChars are the bottleneck though which all changes occur
	if (!readOnly) {
		if (collectingUndo) {
			// Save into the undo/redo stack, but only the characters - not the formatting
			// This takes up about half load time
			char *data = new char[insertLength/2];
			for (int i=0;i<insertLength/2;i++) {
				data[i] = s[i * 2];
			}
			AppendAction(insertAction, position, data, insertLength/2);
		}

		BasicInsertString(position, s, insertLength);
	}
}

void Document::InsertCharStyle(int position, char ch, char style) {
	char s[2];
	s[0] = ch;
	s[1] = style;
	InsertString(position*2, s,2);
}

void Document::SetStyleAt(int position, char style, char mask) {
	SetByteAt(position*2 + 1, (ByteAt(position*2 + 1) & ~mask) | style);
}

void Document::AppendAction(actionType at, int position, char *data, int length) {
	//dprintf("%% %d action %d %d %d\n", at, position, length, currentAction);
	if (currentAction >= 2) {
		// See if current action can be coalesced into previous action
		// Will work if both are inserts or deletes and position is same or two different
		if ((at != actions[currentAction-1].at) || (abs(position-actions[currentAction-1].position) > 2)) {
			currentAction++;
		} else if (currentAction == savePoint) {
			currentAction++;
		}
	} else {
		currentAction++;
	}
	actions[currentAction].Create(at, position, data, length);
	currentAction++;
	actions[currentAction].Create(startAction);
	maxAction = currentAction;
}

void Document::DeleteChars(int position, int deleteLength) {
	// InsertString and DeleteChars are the bottleneck though which all changes occur
	if (!readOnly) {
		if (collectingUndo) {
			// Save into the undo/redo stack, but only the characters - not the formatting
			char *data = new char[deleteLength/2];
			for (int i=0;i<deleteLength/2;i++) {
				data[i] = ByteAt(position + i * 2);
			}
			AppendAction(removeAction, position, data, deleteLength/2);
		}

		BasicDeleteChars(position, deleteLength);
	}
}

int Document::ByteLength() {
	return length;
}

int Document::Length() {
	return ByteLength() / 2;
}

void Document::SetLineCache() {
	// TODO: Maintain cache with every change so does not need to be regenerated.
	if (!lc.valid) {
		lc.SetValue(0,0);
		lc.lines = 1;
		char chPrev = ' ';
		for (int i=0; i < length; i+=2) {
			char ch = ByteAt(i);
			if (ch == '\r') {
				lc.SetValue(lc.lines,i/2+1);
				lc.lines++;
			} else if (ch == '\n') {
				if (chPrev == '\r') {
					// Patch up what was end of line
					lc.SetValue(lc.lines-1,i/2+1);
				} else {
					lc.SetValue(lc.lines,i/2+1);
					lc.lines++;
				}
			}
			chPrev = ch;
		}
		lc.SetValue(lc.lines,length/2);
		lc.valid = true;
	}
}

int Document::Lines() {
	SetLineCache();
	return lc.lines;
}

bool Document::IsReadOnly() {
	return readOnly;
}

void Document::SetReadOnly(bool set) {
	readOnly = set;
}

void Document::SetSavePoint() {
	savePoint = currentAction;
}

bool Document::IsSavePoint() {
	return savePoint == currentAction;
}

// Without undo

void Document::BasicInsertString(int position, char *s, int insertLength) {
	RoomFor(insertLength);
	GapTo(position);
	memcpy(body + part1len, s, insertLength);
	length += insertLength;
	part1len += insertLength;
	gaplen -= insertLength;
	part2body = body + gaplen;
	lc.valid = false;
}

void Document::BasicDeleteChars(int position, int deleteLength) {
	GapTo(position);
	length -= deleteLength;
	gaplen += deleteLength;
	part2body = body + gaplen;
	lc.valid = false;
}

bool Document::SetUndoCollection(bool collectUndo) {
	collectingUndo = collectUndo;
	return collectingUndo;
}

bool Document::IsCollectingUndo() {
	return collectingUndo;
}

void Document::DeleteUndoHistory() {
	for (int i=1;i<maxAction;i++)
		actions[i].Destroy();
	maxAction = 0;
	currentAction = 0;
}

int Document::Undo(int *posEarliestChanged) {
	dprintf("Undoing from %d\n", currentAction);
	int retPosition = 0;	// Where the cursor should be after return
	int changedPosition = 0;	// Earliest byte modified
	if (posEarliestChanged)
		*posEarliestChanged = length;
	if (actions[currentAction].at == startAction && currentAction > 0)
		currentAction--;
	while (actions[currentAction].at != startAction && currentAction > 0) {
		if (actions[currentAction].at == insertAction) {
			BasicDeleteChars(actions[currentAction].position, actions[currentAction].lenData*2);
			retPosition = actions[currentAction].position;
		} else if (actions[currentAction].at == removeAction) {
			char *styledData = new char[actions[currentAction].lenData*2];
			memset(styledData, 0, actions[currentAction].lenData*2);
			for (int i=0;i<actions[currentAction].lenData;i++)
				styledData[i*2] = actions[currentAction].data[i];
			BasicInsertString(actions[currentAction].position, styledData, actions[currentAction].lenData*2);
			delete []styledData;
			retPosition = actions[currentAction].position + actions[currentAction].lenData*2;
		}
		changedPosition = actions[currentAction].position;
		if (posEarliestChanged && (*posEarliestChanged > changedPosition))
			*posEarliestChanged = changedPosition;
		currentAction--;
	}
	return retPosition;
}

int Document::Redo(int *posEarliestChanged) {
	int retPosition = 0;	// Where the cursor should be after return
	int changedPosition = 0;	// Earliest byte modified
	if (posEarliestChanged)
		*posEarliestChanged = length;
	if (actions[currentAction].at == startAction && currentAction < maxAction)
		currentAction++;
	while (actions[currentAction].at != startAction && currentAction < maxAction) {
		if (actions[currentAction].at == insertAction) {
			char *styledData = new char[actions[currentAction].lenData*2];
			memset(styledData, 0, actions[currentAction].lenData*2);
			for (int i=0;i<actions[currentAction].lenData;i++)
				styledData[i*2] = actions[currentAction].data[i];
			BasicInsertString(actions[currentAction].position, styledData, actions[currentAction].lenData*2);
			delete []styledData;
			retPosition = actions[currentAction].position + actions[currentAction].lenData*2;
		} else if (actions[currentAction].at == removeAction) {
			BasicDeleteChars(actions[currentAction].position, actions[currentAction].lenData*2);
			retPosition = actions[currentAction].position;
		}
		changedPosition = actions[currentAction].position;
		if (posEarliestChanged && (*posEarliestChanged > changedPosition))
			*posEarliestChanged = changedPosition;
		currentAction++;
	}
	return retPosition;
}

bool Document::CanUndo() {
	//dprintf("Can Undo?\n");
	return (!readOnly) && ((currentAction > 0) && (maxAction > 0));
}

bool Document::CanRedo() {
	return (!readOnly) && (maxAction > currentAction);
}
